import pymysql

from contextlib import contextmanager
from pymysql.cursors import DictCursor
from schematics.models import Model
from schematics.types import StringType, IntType

LOCAL_STATE_JOINING = 1
LOCAL_STATE_DONOR_DESYNCED = 2
LOCAL_STATE_JOINED = 3
LOCAL_STATE_SYNCED = 4

CLUSTER_STATUS_PRIMARY = 'primary'


class GaleraNode(Model):
    """Contains information about a Galera cluster node."""

    # The hostname or IP address of the cluster node
    host = StringType(required=True)

    # The MySQL port of the cluster node
    port = IntType(required=True, default=3306)

    # The name of the cluster this node belongs to
    cluster_name = StringType(required=True)

    # This stores the value of variable `wsrep_cluster_state_uuid` which shows
    # the cluster state UUID which can be used to determine whether the node
    # is part of the cluster. All nodes in the same cluster should have the
    # same UUID
    cluster_state_uuid = StringType(required=True)

    # This stores the value of the variable `wsrep_cluster_status` which shows
    # the primary status of the cluster component that the node is in. The node
    # should only return a value of Primary. Any other value indicates that the
    # node is part of a nonoperational component.
    cluster_status = StringType(required=True,
                                choices=[CLUSTER_STATUS_PRIMARY])

    # Local state of the node in the cluster
    local_state = IntType(required=True, choices=[LOCAL_STATE_JOINING,
                                                  LOCAL_STATE_DONOR_DESYNCED,
                                                  LOCAL_STATE_JOINED,
                                                  LOCAL_STATE_SYNCED])

    # The MySQL username
    username = StringType()

    # The MySQL password
    password = StringType()

    def refresh_state(self):
        """Refreshes the state of the Galera node. Refreshing the state sets the
        appropriate cluster level identifiers corresponding to the cluster
        the node belongs to. It also updates the state of the node in the
        cluster.

        :return: Returns True when all properties of the node can be
            refreshed based on its current state.
        :rtype: bool
        :raises: ModelValidationError, OperationalError
        """
        with self.get_connection() as conn:
            with conn.cursor() as cursor:
                # Fetch the relevant status variables first
                cursor.execute("SHOW GLOBAL STATUS LIKE 'wsrep%'")
                res = {r['Variable_name'].lower(): r['Value'].lower()
                       for r in cursor.fetchall()}

                self.cluster_state_uuid = res.get('wsrep_cluster_state_uuid')
                self.cluster_status = res.get('wsrep_cluster_status')
                self.local_state = int(res.get('wsrep_local_state'))

                # Now fetch the relevant global variables
                cursor.execute('SELECT @@global.wsrep_cluster_name AS '
                               'wsrep_cluster_name')
                res = cursor.fetchone()
                self.cluster_name = res['wsrep_cluster_name']

        self.validate()

        return True

    @contextmanager
    def get_connection(self):
        """Connect to the Galera cluster node.

        :return: Returns a MySQL connection to the node.
        :rtype: Connection
        """
        db = None
        try:
            db = pymysql.connect(
                host=self.host,
                port=self.port,
                user=self.username,
                passwd=self.password,
                cursorclass=DictCursor
            )

            yield db
        finally:
            if db:
                db.close()

    # To uniquely identify a Galera Node all we need is the host and port
    def __hash__(self):
        return hash('%s__%s' % (self.host, self.port))


class GaleraConfig(Model):
    """A model that defines Galera configuration needed by proxysql-tools."""

    writer_hostgroup_id = StringType(required=True)
    reader_hostgroup_id = StringType(required=True)
    cluster_host = StringType(required=True)
    cluster_port = StringType(required=True)
    cluster_username = StringType(required=True)
    cluster_password = StringType(required=True)
    load_balancing_mode = StringType(required=True)



from schematics.models import Model
from schematics.types import StringType, IntType

BACKEND_STATUS_ONLINE = 'ONLINE'
BACKEND_STATUS_SHUNNED = 'SHUNNED'
BACKEND_STATUS_OFFLINE_SOFT = 'OFFLINE_SOFT'
BACKEND_STATUS_OFFLINE_HARD = 'OFFLINE_HARD'


class ProxySQLMySQLBackend(Model):
    """Contains information about MySQL backend"""
    hostgroup_id = IntType(default=0)

    # The name of the host running MySQL. When using unix domain socket,
    # set this to the socket path
    hostname = StringType(required=True)

    # The port MySQL is listening on. When using unix domain socket,
    # set this to 0

    port = IntType(default=3306, required=True)

    status = StringType(choices=[BACKEND_STATUS_ONLINE, BACKEND_STATUS_SHUNNED,
                                 BACKEND_STATUS_OFFLINE_SOFT,
                                 BACKEND_STATUS_OFFLINE_HARD],
                        default=BACKEND_STATUS_ONLINE)

    weight = IntType(default=1)
    compression = IntType(default=0)
    max_connections = IntType(default=10000)
    max_replication_lag = IntType(default=0)
    use_ssl = IntType(choices=[0, 1], default=0)
    max_latency_ms = IntType(default=0)
    comment = StringType(default='')

    def __hash__(self):
        return hash("%s__%s__%s" % (self.hostgroup_id, self.hostname,
                                    self.port))


class ProxySQLMySQLUser(Model):
    """Contains information about MySQL backend user"""

    # Credentials for connecting to MySQL or ProxySQL instance
    username = StringType(required=True)
    password = StringType(required=True)

    active = IntType(choices=[0, 1], default=1)
    use_ssl = IntType(choices=[0, 1], default=0)

    # If there is no matching rule for the queries sent by this user
    # then the traffic is sent to the specified hostgroup

    default_hostgroup = IntType(default=0)

    # The schema to which the connection should change by default
    default_schema = StringType(default='information_schema')
    schema_locked = IntType(choices=[0, 1], default=0)

    # If this is set for the user with which the MySQL client is connecting
    # to ProxySQL (thus a "frontend" user - see
    # below), transaction started within a hostgroup will remain within
    # that hostgroup regardless of any other rules
    transaction_persistent = IntType(choices=[0, 1], default=0)

    # If set, it bypasses the query processing layer (rewriting, caching)
    # and passes through the query directly as is
    # to the backend server
    fast_forward = IntType(choices=[0, 1], default=0)

    # Note, currently all users need both "frontend" and "backend" set to 1.
    # If set to 1, this (username, password)
    # pair is used for authenticating to the MySQL servers against
    # any hostgroup
    backend = IntType(choices=[0, 1], default=1)

    # If set to 1, this (username, password) pair is used for authenticating
    # to the ProxySQL instance

    frontend = IntType(choices=[0, 1], default=1)

    max_connections = IntType(default=10000)

    def __hash__(self):
        return hash('%s__%s' % (self.username, self.backend))


class ProxySQLConfig(Model):
    """A model that defines ProxySQL configuration needed by proxysql-tools."""

    host = StringType(required=True)
    admin_port = StringType(required=True)
    admin_username = StringType(required=True)
    admin_password = StringType(required=True)
    monitor_username = StringType(required=True)
    monitor_password = StringType(required=True)
    virtual_ip = StringType()
    virtual_netmask = StringType()



from click.testing import CliRunner

from proxysql_tools.cli import main
from proxysql_tools.managers.galera_manager import GaleraManager
from tests.integration.library import proxysql_tools_config


def test__main_command_version_can_be_fetched():
    runner = CliRunner()
    result = runner.invoke(main, ['--version'])
    assert result.exit_code == 0


def test__ping_command_can_be_executed(proxysql_manager, tmpdir):
    config = proxysql_tools_config(proxysql_manager, '127.0.0.1', '3306',
                                   'user', 'pass', 10, 11, 'monitor',
                                   'monitor')
    config_file = str(tmpdir.join('proxysql-tool.cfg'))
    with open(config_file, 'w') as fh:
        config.write(fh)

    runner = CliRunner()
    result = runner.invoke(main, ['--config', config_file, 'ping'])
    assert result.exit_code == 0


def test__galera_register_command_can_register_cluster_with_proxysql(
        percona_xtradb_cluster_node, proxysql_manager, tmpdir):
    hostgroup_writer = 10
    hostgroup_reader = 11

    # To start off there should be no nodes in the hostgroups
    assert len(proxysql_manager.fetch_backends(hostgroup_writer)) == 0
    assert len(proxysql_manager.fetch_backends(hostgroup_reader)) == 0

    galera_man = GaleraManager(
        percona_xtradb_cluster_node.host, percona_xtradb_cluster_node.port,
        percona_xtradb_cluster_node.username,
        percona_xtradb_cluster_node.password
    )
    galera_man.discover_cluster_nodes()

    # Validate that there is one healthy node in the cluster
    assert len(galera_man.nodes) == 1

    # Setup the config object
    config = proxysql_tools_config(proxysql_manager,
                                   percona_xtradb_cluster_node.host,
                                   percona_xtradb_cluster_node.port,
                                   percona_xtradb_cluster_node.username,
                                   percona_xtradb_cluster_node.password,
                                   hostgroup_writer, hostgroup_reader,
                                   'monitor', 'monitor')

    config_file = str(tmpdir.join('proxysql-tool.cfg'))
    with open(config_file, 'w') as fh:
        config.write(fh)

    runner = CliRunner()
    result = runner.invoke(main, ['--config', config_file,
                                  'galera', 'register'])
    assert result.exit_code == 0

    writer_backends = proxysql_manager.fetch_backends(hostgroup_writer)
    reader_backends = proxysql_manager.fetch_backends(hostgroup_reader)

    assert len(writer_backends) == len(reader_backends) == 1
    assert writer_backends[0].hostname == reader_backends[0].hostname
    assert writer_backends[0].port == reader_backends[0].port
    assert writer_backends[0].status == reader_backends[0].status



from proxysql_tools import log
from proxysql_tools.entities.galera import (
    LOCAL_STATE_SYNCED,
    LOCAL_STATE_DONOR_DESYNCED
)
from proxysql_tools.entities.proxysql import BACKEND_STATUS_ONLINE
from proxysql_tools.managers.galera_manager import (
    GaleraManager,
    GaleraNodeNonPrimary,
    GaleraNodeUnknownState
)
from proxysql_tools.managers.proxysql_manager import (
    ProxySQLManager,
    ProxySQLAdminConnectionError
)


def register_cluster_with_proxysql(proxy_cfg, galera_cfg):
    """Register a Galera cluster within ProxySQL. The nodes in the cluster
    will be distributed between writer hostgroup and reader hostgroup.

    :param config: The config object.
    :type config: ConfigParser.ConfigParser
    :param proxy_cfg: The ProxySQL config object.
    :type proxy_cfg: ProxySQLConfig
    :param galera_cfg: The Galera config object.
    :type galera_cfg: GaleraConfig
    :return: Returns True on success, False otherwise.
    :rtype: bool
    """
    hostgroup_writer = galera_cfg.writer_hostgroup_id
    hostgroup_reader = galera_cfg.reader_hostgroup_id

    # We also check that the initial node that is being used to register the
    # cluster with ProxySQL is actually a healthy node and part of the primary
    # component.
    galera_man = GaleraManager(galera_cfg.cluster_host,
                               galera_cfg.cluster_port,
                               galera_cfg.cluster_username,
                               galera_cfg.cluster_password)
    try:
        galera_man.discover_cluster_nodes()
    except GaleraNodeNonPrimary:
        log.error('Cluster node %s:%s used for registration is non-primary.' %
                  (galera_man.host, galera_man.port))
        return False
    except GaleraNodeUnknownState:
        log.error('Cluster node %s:%s used for registration is in unknown '
                  'state.' % (galera_man.host, galera_man.port))
        return False

    # First we try to find nodes in synced state.
    galera_nodes_synced = [n for n in galera_man.nodes
                           if n.local_state == LOCAL_STATE_SYNCED]
    galera_nodes_desynced = [n for n in galera_man.nodes
                             if n.local_state == LOCAL_STATE_DONOR_DESYNCED]

    # If we found no nodes in synced or donor/desynced state then we
    # cannot continue.
    if not galera_nodes_synced and not galera_nodes_desynced:
        log.error('No node found in SYNCED or DESYNCED state.')
        return False

    proxysql_man = ProxySQLManager(proxy_cfg.host,
                                   proxy_cfg.admin_port,
                                   proxy_cfg.admin_username,
                                   proxy_cfg.admin_password,
                                   reload_runtime=False)

    try:
        # We also validate that we can connect to ProxySQL
        proxysql_man.ping()

        # Setup the monitoring user used by ProxySQL to monitor the backends
        setup_proxysql_monitoring_user(proxysql_man,
                                       proxy_cfg.monitor_username,
                                       proxy_cfg.monitor_password)

        # Let's remove all the nodes defined in the hostgroups that are not
        # part of this cluster or are not in desired state.
        if galera_nodes_synced:
            desired_state = LOCAL_STATE_SYNCED
            nodes_list = galera_nodes_synced
        else:
            desired_state = LOCAL_STATE_DONOR_DESYNCED
            nodes_list = galera_nodes_desynced

        # Handle write backends
        writer_backends = deregister_unhealthy_backends(proxysql_man,
                                                        galera_man.nodes,
                                                        hostgroup_writer,
                                                        [desired_state])

        # If there are more than one nodes in the writer hostgroup then we
        # remove all but one.
        if len(writer_backends) > 1:
            log.info('There are %d writers. Removing all but one.',
                     len(writer_backends))
            for backend in writer_backends[1:]:
                proxysql_man.deregister_backend(hostgroup_writer,
                                                backend.hostname,
                                                backend.port)
        elif len(writer_backends) == 0:
            # If there are no backends registered in the writer hostgroup
            # then we register one healthy galera node.
            node = nodes_list[0]
            proxysql_man.register_backend(hostgroup_writer,
                                          node.host, node.port)

        # Fetch the final list of writer backends
        writer_backend = [b for b in
                          proxysql_man.fetch_backends(hostgroup_writer)
                          if b.status == BACKEND_STATUS_ONLINE][0]

        # Now deregister all the unhealthy backends in reader hostgroup
        reader_backends = deregister_unhealthy_backends(proxysql_man,
                                                        galera_man.nodes,
                                                        hostgroup_reader,
                                                        [desired_state])

        # Now we register all of the healthy galera nodes in the
        # reader hostgroup.
        for node in nodes_list:
            if (len(reader_backends) > 0 and
                    node.host == writer_backend.hostname and
                    node.port == writer_backend.port):
                continue

            proxysql_man.register_backend(hostgroup_reader,
                                          node.host, node.port)

        # Now filter healthy backends that are common between writer hostgroup
        # and reader hostgroup
        reader_backends = [b for b in
                           proxysql_man.fetch_backends(hostgroup_reader)
                           if b.status == BACKEND_STATUS_ONLINE]

        # If we have more than one backend registered in the reader hostgroup
        # then we remove the ones that are also present in the writer hostgroup
        if len(reader_backends) > 1:
            for b in reader_backends:
                if (b.hostname == writer_backend.hostname and
                        b.port == writer_backend.port):
                    log.info("There is more than one backend registered "
                             "in the reader hostgroup. We remove the ones "
                             "that are also present in the writer hostgroup")
                    proxysql_man.deregister_backend(hostgroup_reader,
                                                    b.hostname, b.port)

        # TODO: Add user sync functionality that syncs MySQL users with Proxy.
    except ProxySQLAdminConnectionError:
        log.error('ProxySQL connection failed.')
        return False
    finally:
        # Reload the ProxySQL runtime so that it picks up all the changes
        # that have been made so far.
        proxysql_man.reload_runtime()

    return True


def deregister_unhealthy_backends(proxysql_man, galera_nodes, hostgroup_id,
                                  desired_states):
    """Remove backends in a particular hostgroup that are not in the Galera
    cluster or whose state is not in one of the desired states.

    :param proxysql_man: ProxySQL manager corresponding to the ProxySQL
        instance.
    :type proxysql_man: ProxySQLManager
    :param galera_nodes: List of Galera nodes.
    :type galera_nodes: list[GaleraNode]
    :param hostgroup_id: The ID of the ProxySQL hostgroup.
    :type hostgroup_id: int
    :param desired_states: Nodes not in this list of states are considered
        unhealthy.
    :type desired_states: list[str]
    :return: A list of ProxySQL backends that correspond to the Galera nodes
        that are part of the cluster.
    :rtype: list[ProxySQLMySQLBackend]
    """
    backend_list = proxysql_man.fetch_backends(hostgroup_id)
    for backend in backend_list:
        # Find the matching galera node and then see if the node state is
        # synced or donor/desynced. If not one of those two states then we
        # deregister the node from ProxySQL as well.
        backend_found_in_cluster = False
        if backend.status == BACKEND_STATUS_ONLINE:
            for node in galera_nodes:
                if (node.host == backend.hostname and node.port == backend.port
                        and node.local_state in desired_states):
                    backend_found_in_cluster = True
                    break

        if not backend_found_in_cluster:
            log.warning('Backend not found in the cluster: '
                        'hostgroup %d, %s:%d %s',
                        hostgroup_id,
                        backend.hostname, backend.port,
                        backend.status)
            proxysql_man.deregister_backend(hostgroup_id, backend.hostname,
                                            backend.port)

            # Remove the backend from list of writer backends as well.
            backend_list.remove(backend)

    return backend_list


def setup_proxysql_monitoring_user(proxysql_man, monitor_user, monitor_pass):
    """Setup the monitoring user used by ProxySQL to monitor the backends.

    :param proxysql_man: ProxySQL manager corresponding to the
        ProxySQL instance.
    :type proxysql_man: ProxySQLManager
    :param monitor_user: The user used by ProxySQL to monitor the backends.
    :type monitor_user: str
    :param monitor_pass: The password of user used by ProxySQL to monitor
        the backends.
    :type monitor_pass: str
    """
    proxysql_man.set_var('mysql-monitor_username', monitor_user)
    proxysql_man.set_var('mysql-monitor_password', monitor_pass)


def register_mysql_users_with_proxysql():
    pass



from schematics.exceptions import ModelValidationError

from pymysql.err import OperationalError

from proxysql_tools import log
from proxysql_tools.entities.galera import GaleraNode, CLUSTER_STATUS_PRIMARY


class GaleraManager(object):
    def __init__(self, cluster_node_host, cluster_node_port, user, password):
        """Initializes the Galera manager.

        :param cluster_node_host: The Galera cluster host to operate against.
        :type cluster_node_host: str
        :param cluster_node_port: The MySQL port on Galera cluster host to
            connect to.
        :type cluster_node_port: int
        :param user: The MySQL username.
        :type user: str
        :param password: The MySQL password.
        :type password: str
        """
        self.host = cluster_node_host
        self.port = int(cluster_node_port)
        self.user = user
        self.password = password
        self._nodes = []

    @property
    def nodes(self):
        return self._nodes

    def discover_cluster_nodes(self):
        """Given the initial node find all the other nodes in the same cluster.
        It sets up the internal nodes list which is later used to perform
        operations on the nodes or the cluster.

        :return: Returns True on success, False otherwise.
        :rtype: bool
        """
        initial_node = GaleraNode({
            'host': self.host,
            'port': self.port,
            'username': self.user,
            'password': self.password
        })

        # Check that the initial node status is 'PRIMARY'
        self.refresh_and_validate_node_state(initial_node)

        self._nodes = [initial_node]

        with initial_node.get_connection() as conn:
            with conn.cursor() as cursor:
                cursor.execute("SHOW GLOBAL STATUS LIKE "
                               "'wsrep_incoming_addresses'")
                res = {r['Variable_name'].lower(): r['Value'].lower()
                       for r in cursor.fetchall()}

                if not res.get('wsrep_incoming_addresses'):
                    err_msg = ('Node %s:%s unknown status variable '
                               '"wsrep_incoming_addresses"')

                    log.error(err_msg)
                    raise GaleraNodeUnknownState(err_msg)

                log.info('Node %s:%s wsrep_incoming_addresses:: %s' %
                         (initial_node.host, initial_node.port,
                          res['wsrep_incoming_addresses']))

                for host_port in res['wsrep_incoming_addresses'].split(','):
                    host, port = host_port.split(':')

                    # We ignore the initial node from wsrep_incoming_addresses
                    if initial_node.host == host:
                        continue

                    node = GaleraNode({
                        'host': host,
                        'port': int(port),
                        'username': initial_node.username,
                        'password': initial_node.password
                    })

                    # Check that the initial node status is 'PRIMARY'
                    self.refresh_and_validate_node_state(node)

                    if GaleraManager.nodes_in_same_cluster(initial_node, node):
                        self._nodes.append(node)

        return True

    @staticmethod
    def refresh_and_validate_node_state(node):
        """Validates the state of the node to ensure that the node's status
        is PRIMARY and that all the node's properties can be fetched.

        :param node: The object that stores information on the Galera node.
        :type node: GaleraNode
        """
        try:
            node.refresh_state()
            if not node.cluster_status == CLUSTER_STATUS_PRIMARY:
                err_msg = ('Node %s:%s is in non-primary '
                           'state %s' % (node.host, node.port,
                                         node.cluster_status))

                log.error(err_msg)
                raise GaleraNodeNonPrimary(err_msg)
        except (ModelValidationError, OperationalError) as e:
            # The node state cannot be refreshed as some of the
            # properties of the node could not be fetched. We
            # should fail the discovery in such a case as this is
            # unexpected error.
            log.error('Node %s:%s state could not be fetched' %
                      (node.host, node.port))
            raise GaleraNodeUnknownState(e.messages)

    @staticmethod
    def nodes_in_same_cluster(node1, node2):
        """Check to see if the two nodes belong to the same cluster.

        :param node1: The Galera node to be compared.
        :type node1: GaleraNode
        :param node2: The Galera node to be compared.
        :type node2: GaleraNode
        :return: Returns True if both nodes are in the same cluster,
            False otherwise.
        :rtype: bool
        """
        return node1.cluster_state_uuid == node2.cluster_state_uuid


class GaleraNodeUnknownState(Exception):
    pass


class GaleraNodeNonPrimary(Exception):
    pass



from proxysql_tools.entities.galera import GaleraConfig
from proxysql_tools.entities.proxysql import (
    BACKEND_STATUS_ONLINE,
    ProxySQLConfig
)
from proxysql_tools.galera import register_cluster_with_proxysql
from proxysql_tools.managers.galera_manager import GaleraManager
from tests.integration.conftest import PXC_MYSQL_PORT, PXC_ROOT_PASSWORD
from tests.integration.library import wait_for_cluster_nodes_to_become_healthy


def test__register_cluster_with_proxysql_is_idempotent(
        percona_xtradb_cluster_node, proxysql_manager):
    hostgroup_writer = 10
    hostgroup_reader = 11

    # Setup the config objects
    proxy_config = ProxySQLConfig({
        'host': proxysql_manager.host,
        'admin_port': proxysql_manager.port,
        'admin_username': proxysql_manager.user,
        'admin_password': proxysql_manager.password,
        'monitor_username': 'monitor',
        'monitor_password': 'monitor'
    })
    galera_config = GaleraConfig({
        'writer_hostgroup_id': hostgroup_writer,
        'reader_hostgroup_id': hostgroup_reader,
        'cluster_host': percona_xtradb_cluster_node.host,
        'cluster_port': percona_xtradb_cluster_node.port,
        'cluster_username': percona_xtradb_cluster_node.username,
        'cluster_password': percona_xtradb_cluster_node.password,
        'load_balancing_mode': 'singlewriter'
    })

    # Do an initial registration of the cluster
    assert register_cluster_with_proxysql(proxy_config, galera_config)

    writer_backends = proxysql_manager.fetch_backends(hostgroup_writer)
    reader_backends = proxysql_manager.fetch_backends(hostgroup_reader)

    # We try to register again and check that registration is idempotent
    assert register_cluster_with_proxysql(proxy_config, galera_config)

    assert proxysql_manager.fetch_backends(hostgroup_writer)[0].hostname == writer_backends[0].hostname  # NOQA
    assert proxysql_manager.fetch_backends(hostgroup_reader)[0].hostname == reader_backends[0].hostname  # NOQA


def test__register_cluster_with_proxysql_removes_incorrect_nodes(
        percona_xtradb_cluster_three_node,
        percona_xtradb_cluster_node, proxysql_manager):
    hostgroup_writer = 10
    hostgroup_reader = 11

    # Let's wait for our 3-node cluster to become healthy first
    wait_for_cluster_nodes_to_become_healthy(percona_xtradb_cluster_three_node)

    # Setup the config objects
    proxy_config = ProxySQLConfig({
        'host': proxysql_manager.host,
        'admin_port': proxysql_manager.port,
        'admin_username': proxysql_manager.user,
        'admin_password': proxysql_manager.password,
        'monitor_username': 'monitor',
        'monitor_password': 'monitor'
    })
    galera_config = GaleraConfig({
        'writer_hostgroup_id': hostgroup_writer,
        'reader_hostgroup_id': hostgroup_reader,
        'cluster_host': percona_xtradb_cluster_node.host,
        'cluster_port': percona_xtradb_cluster_node.port,
        'cluster_username': percona_xtradb_cluster_node.username,
        'cluster_password': percona_xtradb_cluster_node.password,
        'load_balancing_mode': 'singlewriter'
    })

    # Register the node first from the 1-node cluster
    assert register_cluster_with_proxysql(proxy_config, galera_config)

    # Now we register the nodes from the three node cluster
    galera_man = GaleraManager(percona_xtradb_cluster_three_node[0]['ip'],
                               PXC_MYSQL_PORT, 'root', PXC_ROOT_PASSWORD)

    # Setup the config object again
    galera_config = GaleraConfig({
        'writer_hostgroup_id': hostgroup_writer,
        'reader_hostgroup_id': hostgroup_reader,
        'cluster_host': galera_man.host,
        'cluster_port': galera_man.port,
        'cluster_username': galera_man.user,
        'cluster_password': galera_man.password,
        'load_balancing_mode': 'singlewriter'
    })

    assert register_cluster_with_proxysql(proxy_config, galera_config)

    # There should be one writer backend and two reader backends.
    # Also validate the backends to make sure they are part of the 3-node
    # cluster.
    galera_man.discover_cluster_nodes()

    writer_backends = [b for b in
                       proxysql_manager.fetch_backends(hostgroup_writer)
                       if b.status == BACKEND_STATUS_ONLINE]
    reader_backends = [b for b in
                       proxysql_manager.fetch_backends(hostgroup_reader)
                       if b.status == BACKEND_STATUS_ONLINE]

    assert len(writer_backends) == 1
    assert len(reader_backends) == 2

    writer_nodes = [n for n in galera_man.nodes for w in writer_backends
                    if n.host == w.hostname and n.port == w.port]
    reader_nodes = [n for n in galera_man.nodes for r in reader_backends
                    if r.hostname == n.host and r.port == n.port]

    assert len(writer_nodes) == 1
    assert len(reader_nodes) == 2

    # Now we validate that the writer nodes and readers node form a disjoint
    # set
    assert len(set(writer_nodes).intersection(set(reader_nodes))) == 0



import pytest

from proxysql_tools.entities.proxysql import (
    BACKEND_STATUS_OFFLINE_SOFT,
    BACKEND_STATUS_OFFLINE_HARD,
    ProxySQLMySQLBackend,
    ProxySQLMySQLUser
)
from proxysql_tools.managers.proxysql_manager import (
    ProxySQLManager,
    ProxySQLAdminConnectionError
)


def test__can_connect_to_proxysql_admin_interface(proxysql_manager):
    with proxysql_manager.get_connection() as conn:
        with conn.cursor() as cursor:
            sql = ("SELECT variable_value AS version FROM global_variables "
                   "WHERE variable_name='mysql-server_version'")
            cursor.execute(sql)
            result = cursor.fetchone()

            version = result['version']

    assert version == "5.5.30"


def test__ping_successful_on_successful_connection(proxysql_manager):
    assert proxysql_manager.ping()


def test__ping_raises_exception_on_connection_failure():
    proxysql_man = ProxySQLManager(host='172.0.0.1', port=10000,
                                   user='user_error', password='pass_error')

    with pytest.raises(ProxySQLAdminConnectionError):
        proxysql_man.ping()


def test__can_register_mysql_backend(proxysql_manager):
    backend = get_mysql_backend('db01')

    assert proxysql_manager.register_backend(
        backend.hostgroup_id, backend.hostname, backend.port)

    with proxysql_manager.get_connection() as conn:
        assert proxysql_manager.is_mysql_backend_registered(backend, conn)


def test__fetch_mysql_backends(proxysql_manager):
    backend = get_mysql_backend('db01')

    assert proxysql_manager.register_backend(
        backend.hostgroup_id, backend.hostname, backend.port)

    backends_list = proxysql_manager.fetch_backends()
    assert len(backends_list) == 1
    assert backends_list.pop().hostname == 'db01'


def test__fetch_mysql_backends_belonging_to_hostgroup(proxysql_manager):
    backend = get_mysql_backend('db01')
    assert proxysql_manager.register_backend(
        backend.hostgroup_id, backend.hostname, backend.port)

    backend = get_mysql_backend('db02', hostgroup_id=200)
    assert proxysql_manager.register_backend(
        backend.hostgroup_id, backend.hostname, backend.port)

    backends_list = proxysql_manager.fetch_backends(hostgroup_id=200)
    assert len(backends_list) == 1
    assert backends_list.pop().hostname == 'db02'


def test__update_mysql_backend_status(proxysql_manager):
    backend = get_mysql_backend('db01')

    proxysql_manager.register_backend(
        backend.hostgroup_id, backend.hostname, backend.port)

    assert proxysql_manager.update_mysql_backend_status(
        backend.hostgroup_id, backend.hostname, backend.port,
        BACKEND_STATUS_OFFLINE_SOFT)

    backends_list = proxysql_manager.fetch_backends()
    assert backends_list.pop().status == BACKEND_STATUS_OFFLINE_SOFT


def test__mysql_backend_can_be_deregistered(proxysql_manager):
    backend = get_mysql_backend('db01')

    proxysql_manager.register_backend(
        backend.hostgroup_id, backend.hostname, backend.port)

    assert proxysql_manager.deregister_backend(
        backend.hostgroup_id, backend.hostname, backend.port)
    backends_list = proxysql_manager.fetch_backends()

    assert backends_list.pop().status == BACKEND_STATUS_OFFLINE_HARD


def test__can_register_mysql_user(proxysql_manager):
    user = get_mysql_user('ot')

    assert proxysql_manager.register_mysql_user(user.username, user.password,
                                                user.default_hostgroup)

    with proxysql_manager.get_connection() as conn:
        assert proxysql_manager.is_mysql_user_registered(user, conn)


def test__fetch_mysql_users(proxysql_manager):
    user = get_mysql_user('ot')

    assert proxysql_manager.register_mysql_user(user.username, user.password,
                                                user.default_hostgroup)

    users_list = proxysql_manager.fetch_mysql_users()
    assert len(users_list) == 1
    assert users_list.pop().username == 'ot'


def test__fetch_mysql_users_with_default_hostgroup(proxysql_manager):
    user = get_mysql_user('ot')
    assert proxysql_manager.register_mysql_user(user.username, user.password,
                                                user.default_hostgroup)

    user = get_mysql_user('aleks', default_hostgroup_id=200)
    assert proxysql_manager.register_mysql_user(user.username, user.password,
                                                user.default_hostgroup)

    users_list = proxysql_manager.fetch_mysql_users(default_hostgroup_id=200)
    assert len(users_list) == 1
    assert users_list.pop().username == 'aleks'


def test__can_set_proxysql_variables(proxysql_manager):
    proxysql_manager.set_var('mysql-monitor_username', 'monitor_user')
    proxysql_manager.set_var('mysql-monitor_password', 'monitor_pass')

    proxysql_vars = proxysql_manager.get_vars()
    assert proxysql_vars['mysql-monitor_username'] == 'monitor_user'
    assert proxysql_vars['mysql-monitor_password'] == 'monitor_pass'


def get_mysql_backend(hostname, hostgroup_id=None):
    backend = ProxySQLMySQLBackend()
    backend.hostgroup_id = 10 if hostgroup_id is None else hostgroup_id
    backend.hostname = hostname
    backend.port = 10000

    return backend


def get_mysql_user(username, default_hostgroup_id=None):
    user = ProxySQLMySQLUser()
    user.username = username
    user.password = 'secret_password'
    user.default_hostgroup = 10 if default_hostgroup_id is None else \
        default_hostgroup_id

    return user



import os
from ConfigParser import ConfigParser

import click

from schematics.exceptions import ModelValidationError, ModelConversionError

import proxysql_tools.aws
import proxysql_tools.galera
from proxysql_tools import setup_logging, log, __version__
from proxysql_tools.entities.galera import GaleraConfig
from proxysql_tools.entities.proxysql import ProxySQLConfig
from proxysql_tools.managers.proxysql_manager import (
    ProxySQLManager,
    ProxySQLAdminConnectionError
)

pass_cfg = click.make_pass_decorator(ConfigParser, ensure=True)


@click.group(invoke_without_command=True)
@click.option('--debug', help='Print debug messages', is_flag=True,
              default=False)
@click.option('--config', help='ProxySQL Tools configuration file.',
              default='/etc/twindb/proxysql-tools.cfg',
              show_default=True)
@click.option('--version', help='Show tool version and exit.', is_flag=True,
              default=False)
@pass_cfg
@click.pass_context
def main(ctx, cfg, debug, config, version):
    if ctx.invoked_subcommand is None:
        if version:
            print(__version__)
            exit(0)
        else:
            print(ctx.get_help())
            exit(-1)

    setup_logging(log, debug=debug)

    if os.path.exists(config):
        cfg.read(config)
    else:
        log.error("Config file %s doesn't exist", config)
        exit(1)


@main.command()
@pass_cfg
def ping(cfg):
    """Checks the health of ProxySQL."""
    p_cfg = None

    cfg_opts = {item[0]: item[1] for item in cfg.items('proxysql')}
    try:
        p_cfg = ProxySQLConfig(cfg_opts)
        p_cfg.validate()

        log.debug('Performing health check on ProxySQL instance at %s:%s' %
                  (p_cfg.host, p_cfg.admin_port))

        proxysql_man = ProxySQLManager(host=p_cfg.host,
                                       port=p_cfg.admin_port,
                                       user=p_cfg.admin_username,
                                       password=p_cfg.admin_password)
        proxysql_man.ping()
    except ProxySQLAdminConnectionError:
        log.error('ProxySQL ping failed. Unable to connect at %s:%s '
                  'using username %s and password %s' %
                  (p_cfg.host, p_cfg.admin_port,
                   p_cfg.admin_username, "*" * len(p_cfg.admin_password)))
        exit(1)
    except (ModelValidationError, ModelConversionError) as e:
        log.error('ProxySQL configuration options error: %s' % e)
        exit(1)

    log.info('ProxySQL ping on %s:%s successful.' %
             (cfg_opts['host'], cfg_opts['admin_port']))


@main.group()
@pass_cfg
def aws(cfg):
    """Commands to interact with ProxySQL on AWS."""
    pass


@aws.command()
@pass_cfg
def notify_master(cfg):
    """The notify_master script for keepalived."""
    log.debug('Switching to master role and executing keepalived '
              'notify_master script.')
    proxysql_tools.aws.notify_master(cfg)


@main.group()
@pass_cfg
def galera(cfg):
    """Commands for ProxySQL and Galera integration."""
    pass


@galera.command()
@pass_cfg
def register(cfg):
    """Registers Galera cluster nodes with ProxySQL."""
    proxy_cfg = galera_cfg = None

    proxy_options = {item[0]: item[1] for item in cfg.items('proxysql')}
    try:
        proxy_cfg = ProxySQLConfig(proxy_options)
        proxy_cfg.validate()
    except ModelValidationError as e:
        log.error('ProxySQL configuration options error: %s' % e)
        exit(1)

    galera_options = {item[0]: item[1] for item in cfg.items('galera')}
    try:
        galera_cfg = GaleraConfig(galera_options)
        galera_cfg.validate()
    except (ModelValidationError, ModelConversionError) as e:
        log.error('Galera configuration options error: %s' % e)
        exit(1)

    if not proxysql_tools.galera.register_cluster_with_proxysql(proxy_cfg,
                                                                galera_cfg):
        log.error('Registration of Galera cluster nodes failed.')
        exit(1)

    log.info('Registration of Galera cluster nodes successful.')
